
static final float MIN_COLLISION_EPSILON = 0.0001;
static int lastUsedObjectID = 0;
int wPressed = 'w';
int aPressed = 'a';
int sPressed = 's';
int dPressed = 'd';
int noPressed;
int downPressed = DOWN;
int upPressed = UP;
int rightPressed = RIGHT;
int leftPressed = LEFT;

abstract class GameObject {


  GameWorld gameWorld;
  int objectID;

  GameObject(GameWorld _gameWorld) {
    objectID = lastUsedObjectID++;
    gameWorld = _gameWorld;
  }

  /** default physics: unmovable object, does not do a thing */

  void updateA(float deltaTimeSeconds) {
  }
  void updateV(float deltaTimeSeconds) {
  }
  void updateS(float deltaTimeSeconds) {
  }
  void handleCollisions(float deltaTimeSeconds) {
  }

  abstract void draw();
}

abstract class SphericalGameObject extends GameObject {

  PVector s, v, a; //!< in meters. coordinate system: 0,0 is top-left. Location s, velocity v, acceleration a.
  PVector gravity, gravity_TO_PLAYER;
  float radius; // in meters
  color sphereColor;
  float dampingFactor = 0.15; //!<  > 0; relative amount of velocity retained after 1 second (1.0 == no damping)
  float mass = 0.01;

  SphericalGameObject(GameWorld _gameWorld, float x, float y, float _radius) {
    super(_gameWorld);
    radius = _radius;

    s = new PVector(x, y);
    v = new PVector(0.0, 0.0);
    a = new PVector(0.0, 0.0);

    gravity = new PVector (-0.00098, 0.00098);  //gravity-vector for the bottom-left corner
  }

  void updateA(float deltaTimeSeconds) {
  }

  void updateV(float deltaTimeSeconds) {
    v.add(PVector.mult(a, deltaTimeSeconds));
    v.mult(pow(dampingFactor, deltaTimeSeconds));
  }

  void updateS(float deltaTimeSeconds) {
    s.add(PVector.mult(v, deltaTimeSeconds)).add(PVector.mult(a, -0.5*deltaTimeSeconds*deltaTimeSeconds));
  }

  void handleCollisions(float deltaTimeSeconds) {
    for (GameObject o : gameWorld.gameObjects) {
      if (o instanceof WorldBoundary) {
        handleWorldBoundaryCollision((WorldBoundary) o, deltaTimeSeconds);
      } else if (o instanceof SphericalGameObject) {
        handleSphericalGameObjectCollision((SphericalGameObject) o, deltaTimeSeconds);
      } else if (o instanceof Glitter) {
        handleSphericalGameObjectCollision((Glitter) o, deltaTimeSeconds);
      }
    }
  }

  void handleSphericalGameObjectCollision(SphericalGameObject sgo, float deltaTimeSeconds) {
  }


  //handle Collision with Boundaries
  void handleWorldBoundaryCollision(WorldBoundary wb, float deltaTimeSeconds) {
    if (this instanceof Unicorn) {
      if ((s.y<0.05)||(s.x>gameWorld.worldWidthMeters-0.05)) { 
        float vdotn = PVector.dot(v.normalize().mult(0.5), wb.n);

        PVector vold = v.copy().normalize().mult(0.5);
        v.sub(PVector.mult(wb.n, 3.0 * vdotn));
        s.add(PVector.mult(vold, -deltaTimeSeconds));
        s.add(PVector.mult(v.normalize().mult(0.5), deltaTimeSeconds));
      }
      //saves the unicorn
      if (s.y<0.04) {
        s.y=0.25;
      }
      if (s.x>gameWorld.worldWidthMeters-0.04) {
        s.x=gameWorld.worldWidthMeters-0.25;
      }
    }
  }
}

abstract class Bubble extends SphericalGameObject {
  int counter=0;
  boolean popped;
  float startSpeed;

  Bubble(GameWorld _gameWorld, float x, float y, float _radius, boolean popped) {
    super(_gameWorld, x, y, _radius);
    this.popped=popped;
    this.startSpeed=0.05;
  }
  void updateA(float deltaTimeSeconds) {
    a=new PVector(-0.5*gameWorld.updateSpeed, 0.5*gameWorld.updateSpeed);
  }
  void updateV(float deltaTimeSeconds) {
    gameWorld.speed=startSpeed*gameWorld.updateSpeed;
    a=new PVector(-0.5*gameWorld.updateSpeed, 0.5*gameWorld.updateSpeed);
    a.mult(0.03*gameWorld.updateSpeed);
    v.add(PVector.mult(a, deltaTimeSeconds));

    //limit speed of bubbles
    if (v.mag()>gameWorld.speed) {
      v.setMag(gameWorld.speed);
    }
  }
  void updateS(float deltaTimeSeconds) {
    s.add(PVector.mult(v, deltaTimeSeconds).add(PVector.mult(a, 0.5).mult(deltaTimeSeconds).mult(deltaTimeSeconds)));

    //pulse bubble
    counter++;
    if (counter<=500&&counter>250) {
      radius+=0.0002;
      if (counter==500) {
        counter=0;
      }
    }
    if (counter<=250) {
      radius-=0.0002;
    }
  }
  abstract void draw();
}
class GoodBubble extends Bubble {
  Sweet sweet;

  GoodBubble(GameWorld _gameWorld, float x, float y, float _radius, boolean _popped, Sweet _sweet) {
    super(_gameWorld, x, y, _radius, _popped);
    this.sweet=_sweet;
  }

  void handleSphericalGameObjectCollision(SphericalGameObject sgo, 
    float deltaTimeSeconds) {

    if (! (sgo instanceof Unicorn)||this.popped==true) { // we are only interested in Bubble-Unicorn collisions
      return;
    }
    if (PVector.dist(sgo.s, s) < sgo.radius+radius-0.09) {
      this.popped=true;
      this.sweet.withoutBubble=true;
      gameWorld.bubbleCounter--;


      popBubble.play();  //AUDIO
      popBubble.rewind(); //AUDIO
    }
  }


  void draw() {
    noStroke();
    imageMode(CENTER);
    image(goodBubble, s.x, s.y, radius+0.2, radius+0.2);
    imageMode(CORNER);

    sweet.updatePos(s.x, s.y);
  }
}

class BadBubble extends Bubble {

  BadBubble(GameWorld _gameWorld, float x, float y, float _radius, boolean _popped) {
    super(_gameWorld, x, y, _radius, _popped);
  }

  void handleSphericalGameObjectCollision(SphericalGameObject sgo, 
    float deltaTimeSeconds) {        
    if (! (sgo instanceof Unicorn)||this.popped==true) { // we are only interested in Bubble-Unicorn collisions
      return;
    }
    if (PVector.dist(sgo.s, this.s) < sgo.radius+radius-0.09) {
      if (gameWorld.saveRespawn==false) {
        this.popped=true;
        gameWorld.bubbleCounter--;
        //losing heart when popping a "Bad Bubble"
        if (gameWorld.rainbowMode==false) {
          badBubblePlop.play(); //AUDIO
          badBubblePlop.rewind(); //AUDIO
          gameWorld.lives--;
          lostHeart.play(); //AUDIO
          lostHeart.rewind(); //AUDIO
        } else {
          popBubble.play(); //AUDIO
          popBubble.rewind(); //AUDIO
        }
      }
    }
  }

  void draw() {
    imageMode(CENTER);
    image(badBubble, s.x, s.y, radius+0.2, radius+0.2);
    imageMode(CORNER);
  }
}
class Sweet extends SphericalGameObject {
  float randomIcon= random(0, 3);
  boolean withoutBubble=false;


  Sweet(GameWorld _gameWorld, float x, float y, float _radius) {
    super(_gameWorld, x, y, _radius);
  }

  void updatePos(float x, float y) {//positioned in middle of bubble
    s.x=x-radius/2;
    s.y=y-radius/2;
  }

  void updateA(float deltaTimeSeconds) {
    if (withoutBubble==true) {
      //calculate Vector between Seet and heartPos, then normalize and increase it
      a=new PVector((0.15-s.x),(0.15-s.y)).normalize().mult(4);
    }
  }

  void updateV(float deltaTimeSeconds) {
    if (withoutBubble==true) {
      v.add(PVector.mult(a, deltaTimeSeconds));
      v.mult(pow(dampingFactor, deltaTimeSeconds));
    }
  }

  void updateS(float deltaTimeSeconds) {
    if (withoutBubble==true) {
      s.add(PVector.mult(v, deltaTimeSeconds)).add(PVector.mult(a, -0.5*deltaTimeSeconds*deltaTimeSeconds));
    }
  }
  void draw() {
    //draws random sweet
    if (randomIcon<1) {
      image(lolli, s.x, s.y, radius, radius);
    } else if (randomIcon<2) {
      image(schoki, s.x, s.y, radius, radius);
    } else if (randomIcon<3) {
      image(bonbon, s.x, s.y, radius, radius);
    }
  }
}


// moving Background
class Background extends GameObject {
  PVector s, v, a, startPos;
  float repeatWidth; // when should the image be repeated
  float startSpeed=0.02;

  int alphaPink=255, alphaLightBlue=255, alphaDarkBlue=255;
  int lightBlueHeight=350, darkBlueHeight=1100, starHeight=2000;//when does the background change

  Background(GameWorld _gameWorld, PVector _startPos) {
    super(_gameWorld);
    startPos=_startPos.copy();
    s = startPos.copy();
    v = new PVector(0.0, 0.0);
    a = new PVector(0.0, 0.0);
    repeatWidth = gameWorld.worldHeightMeters;
  }

  void updateV(float deltaTimeSeconds) {
    gameWorld.speed=startSpeed*gameWorld.updateSpeed;//increses speed over time
    v=new PVector(-gameWorld.speed*gameWorld.worldWidthMeters, gameWorld.speed*gameWorld.worldHeightMeters);
  }
  void updateS(float deltaTimeSeconds) {
    s.add(PVector.mult(v, deltaTimeSeconds).add(PVector.mult(a, 0.5).mult(deltaTimeSeconds).mult(deltaTimeSeconds)));
  }

  void draw () { 

    //set background according to gameHeigth
    image(universeSky, s.x, s.y-gameWorld.worldHeightMeters, gameWorld.worldWidthMeters*2, gameWorld.worldHeightMeters*2);
    if (gameWorld.gameHeight>=starHeight) {
      tint(255, alphaDarkBlue);

      if (alphaDarkBlue>=0) {
        alphaDarkBlue--;
      }
    }

    image(skyDarkBlue, s.x, s.y-gameWorld.worldHeightMeters, gameWorld.worldWidthMeters*2, gameWorld.worldHeightMeters*2);

    //fade between day and night music
    if (gameWorld.gameHeight==darkBlueHeight-4) ingame.setGain(-16);
    if (gameWorld.gameHeight==darkBlueHeight-2) ingame.setGain(-22);
    if (gameWorld.gameHeight==darkBlueHeight) ingame.setGain(-28);
    if (gameWorld.gameHeight==darkBlueHeight+2) ingame.setGain(-34);
    if (gameWorld.gameHeight==darkBlueHeight+4) ingame.setGain(-40);

    if (gameWorld.gameHeight==darkBlueHeight) {
      nightSky_ingame.play(); //AUDIO
      nightSky_ingame.loop(); //AUDIO
      if (gameWorld.rainbowMode==true) {
        nightSky_ingame.mute(); //AUDIO
      }
    }
    if (gameWorld.gameHeight>=darkBlueHeight) {
      tint(255, alphaLightBlue);

      if (alphaLightBlue>=0) {
        alphaLightBlue--;
        ingame.pause(); //AUDIO
        ingame.setGain (-10); //AUDIO
        ingame.rewind(); //AUDIO
      }
    }
    image(skyLightBlue, s.x, s.y-gameWorld.worldHeightMeters, gameWorld.worldWidthMeters*2, gameWorld.worldHeightMeters*2);
    noTint();
    if (gameWorld.gameHeight>=lightBlueHeight) {
      tint(255, alphaPink);
      if (alphaPink>=0) {
        alphaPink--;
      }
    }
    image(skyPink, s.x, s.y-gameWorld.worldHeightMeters, gameWorld.worldWidthMeters*2, gameWorld.worldHeightMeters*2);
    noTint();


    //set Image to the beginning
    if (s.y > (startPos.y+repeatWidth)) {
      s = startPos.copy(); //Set imagePos to startPos
    }
  }
}


//Edges of the World :)
class WorldBoundary extends GameObject {

  PVector pA, pB; // points A and B on the Line  
  PVector n; // normal
  float d; // distance origin-plane, for Hessian normal form


  WorldBoundary(GameWorld _gameWorld, PVector _pA, PVector _pB) {
    super(_gameWorld);
    // copy the contents, so that we do not have an external reference into our internal data structures
    pA = _pA.copy();
    pB = _pB.copy();
    PVector u = PVector.sub(pB, pA);
    n = new PVector(u.y, -u.x);
    n.normalize();
    d = PVector.dot(n, pA);
  }

  void draw() {
    image(horizontalClouds, 0, -1.4, gameWorld.worldWidthMeters, gameWorld.worldHeightMeters);
    image(verticalClouds, 2.3, 0, gameWorld.worldWidthMeters, gameWorld.worldHeightMeters);
  }
}
//springs calculated with hooke's law
 
/*credits: to Daniel Shiffman (amazing tutorials on youtube about game physics)
The spring code is based on one of Daniel Shiffmans examples form the book Nature of Code. Also I have read this chapter quite often to understand the physics behind springs.
https://github.com/nature-of-code/noc-examples-processing/tree/master/chp03_oscillation/Exercise_3_16_springs_array
*/

//springs calculated with hooke's law
class Spring extends GameObject {

  // position
  PVector anchor;

  // Rest length and spring constant
  float len;
  float k = 0.18;

  Unicorn a;
  Unicorn b;

  // Constructor
  Spring(GameWorld _gameWorld, Unicorn a_, Unicorn b_, float l) {
    super(_gameWorld);
    a = a_;
    b = b_;
    len = l;
  } 

  // Calculate spring force
  void update(float deltaTimeSeconds) {
    if(gameWorld.rainbowMode==true){
      k=25;
    }else{
      k = 0.18;
    }
    // Vector pointing from one unicorn to the other
    PVector force = PVector.sub(a.s, b.s);
    // What is distance
    float d = force.mag();
    float stretch = d - len;
      // Stretch is difference between current distance and rest length
      // Calculate force according to Hooke's Law
      // F = k * stretch
      force.normalize();
      force.mult(-1 * k * stretch * 0.01);
      a.applyForce(force, deltaTimeSeconds);
      force.mult(-1);
      b.applyForce(force, deltaTimeSeconds);
   }
   
  void draw() {
  }
}

// Glitter is our AI 
// In its normal state it is drawn to the player. Is the player too close to the glitter it moves away

/*Credits to Prof. Fuchs: most of the code is from the example "floyd the boid" in chapter AI 1 in Übungen Games*/

class Glitter extends SphericalGameObject {

  PVector s, v, a; //!< in meters. coordinate system: 0,0 is top-center. Location s, velocity v, acceleration a
  float radius; // in meters
  color glitterColor;
  float mass = 0.1;
  boolean popped;
  PImage glitterImage;

  //behavioral attributes
  PVector destination;//the coordinates, at which glitter is drawn to, random
  float turnRate = 0.5; // in rad/s
  float maxAcceleration = 0.6;
  float repulsionAmount = 2.0;
  float attractionAmount = 2.0;
  float repulsionVisibilityRadius = 1.0; 
  float attractionVisibilityRadius = 2.0; 
  float cosAngleOfMaximalVisibility = -1;
  float directionalInfluence = 0.01; 
  float destinationAcceleration = 1;
  float maxVelocity = 0.35; 

  int currentStatus = 0;
  final int ATTRACTION = 0;
  final int REPULSION = 1;

  PVector alignmentOfNearbyGlitter;

  Glitter(GameWorld _gameWorld, float x, float y, float _radius, PImage _glitterImage) {
    super(_gameWorld, x, y, _radius);
    radius = _radius;
    s = new PVector(x, y);
    float angle = random(2.0*PI); //Random orientation
    v = new PVector(cos(angle)*0.001, sin(angle)*0.001);
    a = new PVector(0.0, 0.0);
    alignmentOfNearbyGlitter = new PVector(0.0, 0.0);

    glitterImage = _glitterImage; 

    destination = new PVector(random(gameWorld.worldWidthMeters), random(gameWorld.worldHeightMeters));
  }
  
  //State change - if glitter gets into REPULSION state it changes color

  void updateColor(int currentStatus) { 
    switch (currentStatus) {

    case ATTRACTION:
      glitterImage = glitter;
      glitterColor = 0;
      break;
    case REPULSION: 
      glitterImage = glitterRepulsion;
      break;
    }
  }

  void updateA(float deltaTimeSeconds) {

    a = PVector.sub(destination, s);
    a.mult(destinationAcceleration);

    //avoid Unicorn destination 
    PVector localDir1 = PVector.sub(gameWorld.player1.s, s);
    float distance1 = PVector.sub(gameWorld.player1.s, s).mag();

    attractionAndRepulsion(localDir1, distance1);

    PVector localDir2 = PVector.sub(gameWorld.player2.s, s);
    float distance2 = PVector.sub(gameWorld.player2.s, s).mag();

    attractionAndRepulsion(localDir2, distance2);
  }
  
  void attractionAndRepulsion(PVector direction, float length) {
    if (length > EPSILON) {   
      direction.mult(1.0/length);
      PVector orientation = direction.copy().normalize();
      float cosBetweenOrientations = orientation.dot(v.copy().normalize());
      if (cosBetweenOrientations >= cosAngleOfMaximalVisibility && length <= repulsionVisibilityRadius) { 
        a.add(PVector.mult(direction, -1 * attractionAmount * pow(length, -4)));
        destination = new PVector(random(gameWorld.worldWidthMeters), random(gameWorld.worldHeightMeters));
        updateColor(REPULSION);
      } else if (cosBetweenOrientations >= cosAngleOfMaximalVisibility && length >= attractionVisibilityRadius) { 
        a.add(PVector.mult(direction, repulsionAmount * pow(length, -2)));
        updateColor(ATTRACTION);
      }
      a.mult(min(1.0, maxAcceleration / a.mag()));
    }
  }

  void updateV(float deltaTimeSeconds) {
    limitTurnRate(PVector.add(v, PVector.mult(a, deltaTimeSeconds)).add(alignmentOfNearbyGlitter.mult(directionalInfluence * v.mag())), deltaTimeSeconds);
    // speed limit
    v.mult(min(1.0, maxVelocity / v.mag()));
  }

  void updateS(float deltaTimeSeconds) {
    s.add(PVector.mult(v, deltaTimeSeconds)).add(PVector.mult(a, -0.5*deltaTimeSeconds*deltaTimeSeconds));
  }

  void computeOtherBoidsAlignment() {
    
    alignmentOfNearbyGlitter.x = 0;
    alignmentOfNearbyGlitter.y = 0;

    for (GameObject o : gameWorld.gameObjects) {
      if (o instanceof Glitter) {
        Glitter b= (Glitter) o;
        PVector localDir = PVector.sub(s, b.s);
        float localLength = localDir.mag();
        if (localLength > EPSILON && localLength <= attractionVisibilityRadius) {
          PVector orientation = localDir.copy().normalize();
          float cosBetweenOrientations = orientation.dot(v.copy().normalize());
          if (cosBetweenOrientations >= cosAngleOfMaximalVisibility) {
            alignmentOfNearbyGlitter.add(b.v.copy().normalize());
          }
        }
      }
    }
  }

// Glitter/ Unicorn collision. 
  void handleSphericalGameObjectCollision(SphericalGameObject sgo, 
    float deltaTimeSeconds) {
    if (! (sgo instanceof Glitter)) { 
      if (! (sgo instanceof Unicorn)||this.popped==true) { // we are only interested in Glitter-Unicorn collisions
        return;
      } else if (PVector.dist(sgo.s, this.s) < (sgo.radius+radius-0.13)&&gameWorld.saveRespawn==false) {
        this.popped = true;
        gameWorld.glitterCounter--;

        if (gameWorld.glitterScore<gameWorld.maxGlitterScore) {
          gameWorld.glitterScore++;
          glitterSound.play(); //AUDIO
          glitterSound.rewind(); //AUDIO
        }
      }
      return;
    }
  }

  void limitTurnRate(PVector newV, float deltaTimeSeconds) {

    PVector desiredOrientation = newV.copy().normalize();
    float desiredLength = newV.mag();

    float desiredOrientationAngle = atan2(desiredOrientation.y, desiredOrientation.x);
    float orientationAngle = atan2(v.y, v.x);
    float delta = desiredOrientationAngle - orientationAngle;

    if (delta > PI) { // the other way is closer
      delta = -2.0 * PI + delta;
    } else if (delta < -PI) {
      delta = 2.0 * PI + delta;
    }

    // limit turn rate
    delta = min(max(delta, -turnRate), turnRate * deltaTimeSeconds);

    orientationAngle += delta;

    v.x = cos(orientationAngle) * desiredLength;
    v.y = sin(orientationAngle) * desiredLength;
  }



  void draw() {

    imageMode(CENTER);
    image(glitterImage, s.x, s.y, radius, radius);
    imageMode(CORNER);
  }
}
